# $Copyright: Copyright (c) 2018 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $

package OpenStackVol;
use LWP::UserAgent;
use JSON;
use Exporter;

@ISA = ('Exporter');
@EXPORT = ('authenticate', 'get_instance_id', 'wait_attaching_detaching_state', 'get_volume_details', 'volume_attach', 'volume_detach', 'initiate_disk_scan', 'is_server_in_systemlist', 'get_server_details', 'server_poweroff', 'wait_powering_off', 'get_native_disk_name', 'update_udid', 'is_vxvm_env', 'include_vxvm_device', 'exclude_vxvm_device','set_global_var','get_timeout_value');

use strict;
use warnings;
use Fcntl qw(:flock);
use POSIX;

use ag_i18n_inc;
my ($http_code, $http_response) = ("", "");

#Attributes and Keywords Name
my ($pOpenStackAuth,$pvirtioscsi) = ("OpenStackAuth","virtio-scsi");

my ($res_name,$start_time) = ("",0);

#Set and Validate commands path
my $psPath = "/usr/bin/ps";
my $catPath = "/usr/bin/cat";
sub ValidateCommandPath {
        my $cmd=shift;
        my $cmdPath = "";

        if(-e "/usr/bin/$cmd") {
                $cmdPath = "/usr/bin/$cmd";
        }
        elsif( -e "/bin/$cmd") {
                $cmdPath = "/bin/$cmd";
        }
        elsif( -e "/usr/local/bin/$cmd") {
                $cmdPath = "/usr/local/bin/$cmd";
        }
        else {
		$cmdPath = "/usr/bin/$cmd";
		VCSAG_LOG_MSG("E", "Command $cmdPath not found",20027,"$cmdPath");
                exit 0;
        }
        return $cmdPath;
}

$psPath = ValidateCommandPath('ps');
$catPath = ValidateCommandPath('cat');

#Read the OpenStack Auth file and get token,compute & cinder public end points
#This sub routine called at the start of each entry point

sub authenticate {
	my ($auth_file) ="$VCSLOG/lock/.OpenStackAuth";
	my ($auth_lockfile) ="$VCSLOG/lock/.OpenStackAuth.lck";
	if (! -e $auth_file) { 
		VCSAG_LOG_MSG("E", "Authentication information not present, $pOpenStackAuth resource should be online",20002,"$pOpenStackAuth"); 
		return;
	}	
	#Acquiring the shared lock
	my ($lock_filehandle) = AcquireSharedLock($auth_lockfile);
	open(my $authfilehandle,"<","$auth_file");
	my $readInfo = <$authfilehandle>;
	close($authfilehandle);
	ReleaseLock($lock_filehandle);
	my @readInfoArry=split(":;",$readInfo);
        my $enc_token = $readInfoArry[0];
        my $token = get_decrypt_token($enc_token);
            
	my $compute_url = $readInfoArry[3];
	my $cinder_url = $readInfoArry[4];
	return ($token, $compute_url, $cinder_url);
}
	
#Get instance id by querying instance meta data 
sub get_instance_id {
        my $res = _cs_request("http://169.254.169.254/openstack/latest/meta_data.json", 'GET', "", "");
	my $instance_id = "";
        if ($res->is_success) {
                my $response = $res->decoded_content;
                my $data = decode_json($response);
                $instance_id = $data->{uuid};
        }
        else {
                #$http_code = $res->code;
                #$http_response = $res->content;
                my $cmd_dmidecode = "dmidecode";
		VCSAG_LOG_MSG ("W", "Could not get instance id using metadata service, trying to get using $cmd_dmidecode command",20012,"$cmd_dmidecode");
		$instance_id = `$cmd_dmidecode -s system-uuid 2>/dev/null`;
		if ($? != 0) {
			my $file_product_uuid = "/sys/class/dmi/id/product_uuid";
			VCSAG_LOG_MSG ("W","Failed to get instance id using $cmd_dmidecode command, trying to get from the file $file_product_uuid",20032,"$cmd_dmidecode","$file_product_uuid");
			$instance_id = `$catPath $file_product_uuid 2>/dev/null`;
			if ($? != 0) {
				 VCSAG_LOG_MSG ("E","Failed to get instance id, $file_product_uuid file not found",20033,"$file_product_uuid");				
			}
		}
		$instance_id = lc($instance_id);
        }
	return $instance_id;
}


#Check if server is part of the system list of the InfoScale Service Group
#return 1 : Server is part of the system list
#return 0 : Server is not a part of the system list

sub is_server_in_systemlist { 
	my ($server_id, $res_name) = @_;
	my $groupName = `$HARES -value $res_name Group`;
	$groupName =~ s/^\s+|\s+$//g;
	my $sysList = `$HAGRP -value $groupName SystemList`;
	$sysList =~ s/^\s+|\s+$//g;
	my @systemList = split /\s+/,$sysList;
	foreach my $system (@systemList) {
		next if ($system =~ /^\d+$/);
		my $instance_id = `$HARES -value $res_name InstanceId $system`;
		$instance_id =~ s/^\s+|\s+$//g;
		if (lc $instance_id eq lc $server_id){
			return 1;
		}
	}
	return 0;
}


#Get volume details of the given volume
sub get_volume_details {
	my ($cinder_url, $volume_id, $token) = @_;
	my $url_volume_details = $cinder_url . "/volumes/" . $volume_id;
	my $res = _cs_request($url_volume_details, 'GET', $token, "");
	if ($res->is_success) {
		my $response = $res->decoded_content;
		my $data = decode_json($response);
		my $status = $data->{volume}{status};
		my $multiattach = $data->{volume}{multiattach};
		my $bootable = $data->{volume}{bootable};
		my @attachments = @{$data->{volume}{attachments}};
		my $server_id = "";
		if (scalar @attachments > 0) {
			$server_id = $attachments[0]->{server_id};
		}
		return ($status, $server_id, $multiattach, $bootable);
	}
	else {
		$http_code = $res->code;
		$http_response = $res->content;
		VCSAG_LOG_MSG ("E", "Failed to get volume details for $volume_id [$http_code] [$http_response]", 20013, $volume_id, $http_code, $http_response);
		return;
	}
}

#Get server details of the given server
sub get_server_details {
	my ($compute_url, $server_id, $token) = @_;
	my $url_server_details = $compute_url . "/servers/" . $server_id;
	my $res = _cs_request($url_server_details, 'GET', $token, "");
	if ($res->is_success) {
		my $response = $res->decoded_content;
		my $data = decode_json($response);
		my $server_status = $data->{server}{status};
		my $task_state = $data->{server}{"OS-EXT-STS:task_state"};
		if (!defined $task_state) { 
			$task_state = "";
		}
		return ($server_status, $task_state);
	}
	else {
		$http_code = $res->code;
		$http_response = $res->content;
		VCSAG_LOG_MSG ("E", "Failed to get server details for $server_id [$http_code] [$http_response]", 20014, $server_id, $http_code, $http_response);
		return;
	}
}

#Power off the given server
sub server_poweroff {
	my ($compute_url, $server_id, $token) = @_;
	VCSAG_LOGDBG_MSG (1, "Powering off the VM $server_id,it may be panicked or stucked",20031,"$server_id");
	my $url_server_poweroff = $compute_url . "/servers/" . $server_id . "/action";
	my $data = {"os-stop" => "null"};
	my $json = encode_json($data);
	my $res = _cs_request($url_server_poweroff, 'POST', $token, $json);
	if ($res->is_success) {
		return;
	}
	else {
		$http_code = $res->code;
		$http_response = $res->content;
		VCSAG_LOG_MSG ("E", "Failed to poweroff the server $server_id [$http_code] [$http_response]", 20015, $server_id, $http_code, $http_response);
		return;
	}
}

#Wait till server gets powered off
sub wait_powering_off {
	my ($compute_url, $server_id, $token) = @_;

	my $sleep_time = 4;
	my ($flag,$ret_count) = get_timeout_value($sleep_time);

	while ($flag or ($ret_count > 0)) {
		if ($flag == 0) {
			$ret_count = $ret_count - 1;
		}

		my @server_details = get_server_details($compute_url, $server_id, $token);
		my $task_state = $server_details[1];
		chomp ($task_state);
		if ($task_state eq "powering-off") {
			sleep $sleep_time;
		} else {
			last;
		}
	}
}

#Attach the volume to the instance
#return 1 : Success
#return 0 : Failed to attach
sub volume_attach { 
	my ($compute_url, $instance_id, $volume_id, $token) = @_;
	my $url_attach = $compute_url . "/servers/" . $instance_id . "/os-volume_attachments";
	my $data = {volumeAttachment => {volumeId => "$volume_id"}};
	my $json = encode_json($data);
	my $res = _cs_request($url_attach, 'POST', $token, $json);
	if ($res->is_success) { 
		return 1;
	}
	else {
		$http_code = $res->code;
		$http_response = $res->content;
		VCSAG_LOG_MSG ("E", "Failed to attach volume $volume_id [$http_code] [$http_response]", 20016, $http_code, $http_response);
		return 0;
	}
}

#Detach the volume from the instance
sub volume_detach {
	my ($compute_url, $instance_id, $volume_id, $token) = @_;
	my $url_detach = $compute_url . "/servers/" . $instance_id . "/os-volume_attachments/" . $volume_id;
	my $res = _cs_request($url_detach, 'DELETE', $token, "");
	if ($res->is_success) {
		return;
	}
	else {
		$http_code = $res->code;
		$http_response = $res->content;
		VCSAG_LOG_MSG ("E", "Failed to detach volume $volume_id [$http_code] [$http_response]", 20017, $http_code, $http_response);
		if ($http_code == 409) {
			sleep 30;
			volume_detach($compute_url, $instance_id, $volume_id, $token);
		}
		return;
	}
}

#Wait till attaching/detaching of the volume from/to the instance gets completed
sub wait_attaching_detaching_state { 
	my ($cinder_url, $volume_id, $token) = @_;
	my $sleep_time = 3;
	my ($flag,$ret_count) = get_timeout_value($sleep_time);

	while ($flag or ($ret_count > 0)) {
		if ($flag == 0) {
			$ret_count = $ret_count - 1;
		}

		my @volume_details = get_volume_details($cinder_url, $volume_id, $token);
		if (@volume_details) {
			my $volume_status = $volume_details[0];
			chomp ($volume_status);
			if ($volume_status =~ /attaching|detaching/) {
				sleep $sleep_time;
			} else {
				last;
			}
		} else {
			VCSAG_LOG_MSG("E", "Unable to fetch volume details for volume id $volume_id",20005,"$volume_id");
			exit 0;
		}
	}
}

#Initiate the VxVM disk scan for any disks/volumes that were newly added
sub initiate_disk_scan {
	sleep 3;
	my $vxdisk = "/opt/VRTS/bin/vxdisk";
	`$vxdisk scandisks > /dev/null 2>&1`;
        my $ret = $?;
        return($ret >> 8);
}

#Check for VxVM env
#return 1 : VxVM libraries are present
#return 0 : VxVM libraries are not present

sub is_vxvm_env {
	my $vxdisk = "/opt/VRTS/bin/vxdisk";
	if (-x $vxdisk) {
		return 1;
	}
	else {
		return 0;
	}
}


sub get_scsi_devices {
	my $SYS_BLOCK = '/sys/block';
	my $serial_id = "";
	my %device_hash = ();
	if (!opendir(DIR, $SYS_BLOCK)) {
		VCSAG_LOG_MSG ("E", "Could not open the directory $SYS_BLOCK",20019,$SYS_BLOCK);
	}
	my @device_list = readdir(DIR);
	closedir(DIR);
	foreach my $device (@device_list) {
		if ($device =~ /^vd.*/) {
			$serial_id = `$catPath /sys/block/$device/serial`;
			if ($? != 0) {
				VCSAG_LOGDBG_MSG (1, "Failed to get serial id for the virtio devices",20020);
				next;
			}
			chomp ($serial_id);
			$device_hash{$serial_id} = $device;
		}
 		elsif ($device =~ /^sd.*/) {
			my $lsblkPath = "/usr/bin/lsblk";
			$lsblkPath = ValidateCommandPath('lsblk');
			my $cmd = "$lsblkPath -nl " . "/dev/" . $device . " -o serial";
			$serial_id = `$cmd`;
			if ($? != 0) {
				VCSAG_LOGDBG_MSG (1, "Failed to get serial id for the $pvirtioscsi devices",20021,"$pvirtioscsi");
				next;
			}
			chomp ($serial_id);
			$device_hash{$serial_id} = $device;		
		}
	}
	return \%device_hash;
}

#Get the native disk name of the volume
sub get_native_disk_name {
	my ($volume_id) = @_;
	my $native_disk_name = "";
	my $short_volume_id = substr($volume_id, 0, 20);
	my $device_hash = get_scsi_devices();
	$native_disk_name = $device_hash->{$short_volume_id};
	return $native_disk_name;
}

#Enable the device for VxVM use/operation
sub include_vxvm_device {
	my ($native_disk) = @_;
	my $cmd = "/opt/VRTS/bin/vxdmpadm include dmpnodename=" . $native_disk;
	`$cmd 2>&1`;
	my $ret = $?;
	return($ret >> 8);
}

#Disable the device for VxVM use/operation
sub exclude_vxvm_device {
	my ($native_disk) = @_;
	my $cmd = "/opt/VRTS/bin/vxdmpadm exclude dmpnodename=" . $native_disk;
	`$cmd 2>&1`;
}

#Clear the udid_mismatch and clone_disk flags if the disk is wrongly marked
sub update_udid {
	my ($native_disk) = @_;
	my $cmd = "/opt/VRTS/bin/vxdisk " . "-c " . "updateudid " . $native_disk;
	`$cmd 2>&1`;
}

sub _cs_request {
	my ($url, $method, $token, $json) = @_;
	my $res = _do_request($url, $method, $token, $json);
	if ($res->code == 401){
		sleep 20;
		my @tokens = authenticate();
		$token = $tokens[0];
		$res = _do_request($url, $method, $token, $json);
	}
	return $res;
}

sub _do_request {
	my ($url, $method, $token, $json) = @_;
	my $req = HTTP::Request->new($method =>$url);
	$req->header('content-type' => 'application/json');
	if ($token ne ""){
		$req->header('x-auth-token' => $token);
	}
	if ($json ne ""){
		$req->content($json);
	}
	my $ua = LWP::UserAgent->new;
	my $res = $ua->request($req);
	return $res;
}

#Acquire exclusive Lock on the file before any updation
sub AcquireLock {

	my ($lock_file) = (@_);
	local *LHANDLE;
	my ($arg, $ret_val) = ("", 0);

	open (LHANDLE, ">$lock_file") or die "could not open file";
	$arg = pack ("SSLLLLLLLL", F_WRLCK (), 0, 0, 0, 0, 0, 0, 0, 0, 0);
	$ret_val = fcntl(LHANDLE, F_SETLKW (), $arg);

	if ($ret_val < 0) {
		VCSAG_LOG_MSG ("E", "Could not get lock on file $lock_file", 20022, "$lock_file");
		exit 0;
	}
	return *LHANDLE;
}

#Acquire the shared lock
sub AcquireSharedLock {
	my ($lock_file) = (@_);
	my $flock_struct = "s s l l i";
	my $buf = pack($flock_struct, F_RDLCK, SEEK_SET, 0, 0, 0);
	
	open(FILE, $lock_file) or die "could not open file";
	$| = 1;
	fcntl(FILE, F_SETLKW, $buf) or die "fcntl(F_SETLK): $!";
	return *FILE;
}

#Release lock from the file after updation
sub ReleaseLock {
	close ($_[0]);
}

sub get_decrypt_token{
        my $enc_pword = shift;
        my  $token = "";

        $enc_pword =~s/^\s+//;
        $enc_pword =~s/\s+$//;
        if ($enc_pword) {
                $token = `$HADECRYPT $enc_pword`;
        }
        return $token;
}

sub set_global_var {
	($res_name,$start_time) = @_;
}

sub get_timeout_value {
 	my $sleep_time = shift;

	my $sEntryPointName = ();
	( $sEntryPointName = $0 ) =~ s#.*/##; # Basename

	my $sEntryPoint_Timeout;

	if($sEntryPointName =~ /online/) {
		$sEntryPoint_Timeout = `$HATYPE -value OpenStackVol OnlineTimeout`;
	} elsif($sEntryPointName =~ /offline/) {
		$sEntryPoint_Timeout = `$HATYPE -value OpenStackVol OfflineTimeout`;
	} elsif($sEntryPointName =~ /clean/) {
		$sEntryPoint_Timeout = `$HATYPE -value OpenStackVol CleanTimeout`;
	} else {
		return (1, -1);
	}

	my $start_time = $start_time;
	my $cur_time = time();
	my $remaining_time = ($sEntryPoint_Timeout )  - ($cur_time - $start_time);
	my $count = int($remaining_time / ($sleep_time + 1));

	return (0,$count);
}

1;
